package ziputils

import java.nio.ByteBuffer
import java.nio.ByteOrder

/**
 * Represents a partial ZIP64 end of central directory record.
 */
internal class EndOfCentralDirectoryRecord64(
    val centralDirectoryOffset: ULong,
) {
    companion object {
        const val SIGNATURE = 0x06064b50U
        const val SIZE = 56

        /**
         * Create `EndOfCentralDirectoryRecord64` from raw byte data.
         * @throws InvalidDataException Provided `ByteArray` is not a supported EOCD.
         * @param data Raw byte data.
         * @param offset Skip first <offset> bytes in data array.
         * @return A `EndOfCentralDirectoryRecord64`.
         */
        @Throws(InvalidDataException::class)
        fun fromByteArray(
            data: ByteArray,
            offset: Int,
        ): EndOfCentralDirectoryRecord64 {
            if (data.size - offset < SIZE) {
                throw InvalidDataException("EOCD64 must be at least 56 bytes")
            }
            val buf = ByteBuffer.wrap(data, offset, SIZE).order(ByteOrder.LITTLE_ENDIAN)
            if (buf.getInt().toUInt() != SIGNATURE) {
                throw InvalidDataException("Invalid signature")
            }
            buf.position(offset + 48)
            return EndOfCentralDirectoryRecord64(
                centralDirectoryOffset = buf.getLong().toULong(),
            )
        }
    }
}